#!C:\bld\obspy_1539460504183\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'obspy==1.1.0','console_scripts','obspy-reftek-rescue'
__requires__ = 'obspy==1.1.0'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('obspy==1.1.0', 'console_scripts', 'obspy-reftek-rescue')()
    )
