# -*- coding: utf-8 -*-
"""
obspy.io.arclink - Arclink inventory files read support for ObsPy
=================================================================

This module provides read support for Arclink inventory files.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
