



python -c "from obspy import read;"
IF %ERRORLEVEL% NEQ 0 exit /B 1
obspy-print -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
obspy-plot -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
obspy-runtests -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
obspy-scan -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
